'''
Created on Jan 24, 2016

@author: nicolas
'''

import os.path
from contextlib import contextmanager
import shutil
import threading
import traceback
from typing import Optional

import six

from lemoncheesecake.consts import ATTACHMENTS_DIR, \
    LOG_LEVEL_DEBUG, LOG_LEVEL_ERROR, LOG_LEVEL_INFO, LOG_LEVEL_WARN
from lemoncheesecake.reporting import Report, ReportLocation
from lemoncheesecake import events
from lemoncheesecake.exceptions import ProgrammingError
from lemoncheesecake.fixture import ScheduledFixtures
from lemoncheesecake.helpers.typecheck import check_type_string, check_type_bool

_session = None  # type: Optional[Session]

_scheduled_fixtures = None  # type: Optional[ScheduledFixtures]


def initialize_session(event_manager, report_dir, report):
    global _session
    _session = Session(event_manager, report_dir, report)
    event_manager.add_listener(_session)


def initialize_fixture_cache(scheduled_fixtures):
    global _scheduled_fixtures
    _scheduled_fixtures = scheduled_fixtures


class _Cursor(object):
    def __init__(self, location, step=None):
        self.location = location
        self.step = step
        self.pending_events = []


class Session(object):
    def __init__(self, event_manager, report_dir, report):
        self.event_manager = event_manager
        self.report_dir = report_dir
        self.report = report
        self.attachments_dir = os.path.join(self.report_dir, ATTACHMENTS_DIR)
        self.attachment_count = 0
        self._attachment_lock = threading.Lock()
        self._failures = set()
        self._local = threading.local()

    @property
    def cursor(self):  # type: () -> _Cursor
        return self._local.cursor

    @cursor.setter
    def cursor(self, cursor):
        self._local.cursor = cursor

    def _hold_event(self, event):
        self.cursor.pending_events.append(event)

    def _flush_pending_events(self):
        for event in self.cursor.pending_events:
            self.event_manager.fire(event)
        del self.cursor.pending_events[:]

    def _discard_pending_event_if_any(self, event_class):
        if self.cursor.pending_events and isinstance(self.cursor.pending_events[-1], event_class):
            self.cursor.pending_events.pop()
            return True
        else:
            return False

    def _discard_or_fire_event(self, event_class, event):
        discarded = self._discard_pending_event_if_any(event_class)
        if not discarded:
            self.event_manager.fire(event)

    def prepare_cursor_for_new_thread(self):
        # In case of a new thread, we have to flush the pending events to avoid sharing the same pending events
        # between multiple threads. It has the minor side-effect of possibly leading to empty step
        self._flush_pending_events()
        return _Cursor(self.cursor.location, self.cursor.step)

    def _mark_location_as_failed(self, location):
        self._failures.add(location)

    def is_successful(self, location=None):
        if location:
            return location not in self._failures
        else:
            return len(self._failures) == 0

    def set_step(self, description, detached=False):
        self._discard_pending_event_if_any(events.StepEvent)
        self.cursor.step = description
        self._hold_event(events.StepEvent(self.cursor.location, description, detached=detached))

    def end_step(self, step):
        self._discard_or_fire_event(events.StepEvent, events.StepEndEvent(self.cursor.location, step))

    def log(self, level, content):
        self._flush_pending_events()
        if level == LOG_LEVEL_ERROR:
            self._mark_location_as_failed(self.cursor.location)
        self.event_manager.fire(events.LogEvent(self.cursor.location, self.cursor.step, level, content))

    def log_check(self, description, is_successful, details):
        self._flush_pending_events()
        if is_successful is False:
            self._mark_location_as_failed(self.cursor.location)
        self.event_manager.fire(
            events.CheckEvent(self.cursor.location, self.cursor.step, description, is_successful, details)
        )

    def log_url(self, url, description):
        self._flush_pending_events()
        self.event_manager.fire(
            events.LogUrlEvent(self.cursor.location, self.cursor.step, url, description)
        )

    @contextmanager
    def prepare_attachment(self, filename, description, as_image=False):
        with self._attachment_lock:
            attachment_filename = "%04d_%s" % (self.attachment_count + 1, filename)
            self.attachment_count += 1
            if not os.path.exists(self.attachments_dir):
                os.mkdir(self.attachments_dir)

        yield os.path.join(self.attachments_dir, attachment_filename)

        self._flush_pending_events()
        self.event_manager.fire(events.LogAttachmentEvent(
            self.cursor.location, self.cursor.step, "%s/%s" % (ATTACHMENTS_DIR, attachment_filename),
            description, as_image
        ))

    def start_test_session(self):
        self.event_manager.fire(events.TestSessionStartEvent(self.report))

    def end_test_session(self):
        self.event_manager.fire(events.TestSessionEndEvent(self.report))

    def start_test_session_setup(self):
        self.cursor = _Cursor(ReportLocation.in_test_session_setup())
        self._hold_event(events.TestSessionSetupStartEvent())

    def end_test_session_setup(self):
        self._discard_pending_event_if_any(events.StepEvent)
        self._discard_or_fire_event(events.TestSessionSetupStartEvent, events.TestSessionSetupEndEvent())

    def start_test_session_teardown(self):
        self.cursor = _Cursor(ReportLocation.in_test_session_teardown())
        self._hold_event(events.TestSessionTeardownStartEvent())

    def end_test_session_teardown(self):
        self._discard_pending_event_if_any(events.StepEvent)
        self._discard_or_fire_event(events.TestSessionTeardownStartEvent, events.TestSessionTeardownEndEvent())

    def start_suite(self, suite):
        self.event_manager.fire(events.SuiteStartEvent(suite))

    def end_suite(self, suite):
        self.event_manager.fire(events.SuiteEndEvent(suite))

    def start_suite_setup(self, suite):
        self.cursor = _Cursor(ReportLocation.in_suite_setup(suite))
        self._hold_event(events.SuiteSetupStartEvent(suite))

    def end_suite_setup(self, suite):
        self._discard_pending_event_if_any(events.StepEvent)
        self._discard_or_fire_event(events.SuiteSetupStartEvent, events.SuiteSetupEndEvent(suite))

    def start_suite_teardown(self, suite):
        self.cursor = _Cursor(ReportLocation.in_suite_teardown(suite))
        self._hold_event(events.SuiteTeardownStartEvent(suite))

    def end_suite_teardown(self, suite):
        self._discard_pending_event_if_any(events.StepEvent)
        self._discard_or_fire_event(events.SuiteTeardownStartEvent, events.SuiteTeardownEndEvent(suite))

    def start_test(self, test):
        self.event_manager.fire(events.TestStartEvent(test))
        self.cursor = _Cursor(ReportLocation.in_test(test))

    def end_test(self, test):
        self.event_manager.fire(events.TestEndEvent(test))

    def skip_test(self, test, reason):
        self.event_manager.fire(events.TestSkippedEvent(test, reason))
        self._mark_location_as_failed(ReportLocation.in_test(test))

    def disable_test(self, test, reason):
        self.event_manager.fire(events.TestDisabledEvent(test, reason))


def get_session():
    # type: () -> Session
    if not _session:
        raise ProgrammingError("Runtime is not initialized")
    return _session


def get_report():
    # type: () -> Report
    report = get_session().report
    if not report:
        raise ProgrammingError("Report is not available")
    return report


def set_step(description, detached=False):
    # type: (str, bool) -> None
    """
    Set a new step.

    :param description: the step description
    :param detached: whether or not the step is "detached"
    """
    check_type_string("description", description)
    get_session().set_step(description, detached=detached)


def end_step(step):
    """
    End a detached step.
    """
    get_session().end_step(step)


@contextmanager
def detached_step(description):
    """
    Context manager. Like set_step, but ends the step at the end of the "with" block.
    Intended to be use with code run through lcc.Thread.
    """
    set_step(description, detached=True)
    try:
        yield
    except Exception:
        # FIXME: use exception instead of last implicit stacktrace
        stacktrace = traceback.format_exc()
        if six.PY2:
            stacktrace = stacktrace.decode("utf-8", "replace")
        log_error("Caught unexpected exception while running test: " + stacktrace)
    end_step(description)


def _log(level, content):
    check_type_string("content", content)
    get_session().log(level, content)


def log_debug(content):
    # type: (str) -> None
    """
    Log a debug level message.
    """
    _log(LOG_LEVEL_DEBUG, content)


def log_info(content):
    # type: (str) -> None
    """
    Log a info level message.
    """
    _log(LOG_LEVEL_INFO, content)


def log_warning(content):
    # type: (str) -> None
    """
    Log a warning level message.
    """
    _log(LOG_LEVEL_WARN, content)


def log_error(content):
    # type: (str) -> None
    """
    Log an error level message.
    """
    _log(LOG_LEVEL_ERROR, content)


def log_check(description, is_successful, details=None):
    check_type_string("description", description)
    check_type_bool("is_successful", is_successful)
    check_type_string("details", details, optional=True)

    get_session().log_check(description, is_successful, details)


def _prepare_attachment(filename, description=None, as_image=False):
    check_type_string("description", description, optional=True)
    check_type_bool("as_image", as_image)

    return get_session().prepare_attachment(filename, description or filename, as_image=as_image)


def prepare_attachment(filename, description=None):
    """
    Context manager. Prepare a attachment using a pseudo filename and an optional description.
    It returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return _prepare_attachment(filename, description)


def prepare_image_attachment(filename, description=None):
    """
    Context manager. Prepare an image attachment using a pseudo filename and an optional description.
    The function returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return _prepare_attachment(filename, description, as_image=True)


def _save_attachment_file(filename, description=None, as_image=False):
    with _prepare_attachment(os.path.basename(filename), description, as_image=as_image) as report_attachment_path:
        shutil.copy(filename, report_attachment_path)


def save_attachment_file(filename, description=None):
    """
    Save an attachment using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    return _save_attachment_file(filename, description)


def save_image_file(filename, description=None):
    """
    Save an image using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    return _save_attachment_file(filename, description, as_image=True)


def _save_attachment_content(content, filename, description=None, as_image=False):
    if type(content) is six.text_type:
        opening_mode = "w"
        if six.PY2:
            content = content.encode("utf-8")
    else:
        opening_mode = "wb"

    with _prepare_attachment(filename, description, as_image=as_image) as path:
        with open(path, opening_mode) as fh:
            fh.write(content)


def save_attachment_content(content, filename, description=None):
    """
    Save a given content as attachment using pseudo filename and optional description.
    """
    return _save_attachment_content(content, filename, description, as_image=False)


def save_image_content(content, filename, description=None):
    """
    Save a given image content as attachment using pseudo filename and optional description.
    """
    return _save_attachment_content(content, filename, description, as_image=True)


def log_url(url, description=None):
    # type: (str, Optional[str]) -> None
    """
    Log an URL.
    """
    check_type_string("url", url)
    check_type_string("description", description, optional=True)

    get_session().log_url(url, description or url)


def get_fixture(name):
    """
    Return the corresponding fixture value. Only fixtures whose scope is pre_run can be retrieved.
    """
    global _scheduled_fixtures

    if not _scheduled_fixtures:
        raise ProgrammingError("Fixture cache has not yet been initialized")

    if not _scheduled_fixtures.has_fixture(name):
        raise ProgrammingError("Fixture '%s' either does not exist or don't have a prerun_session scope" % name)

    try:
        return _scheduled_fixtures.get_fixture_result(name)
    except AssertionError as excp:
        raise ProgrammingError(str(excp))


def add_report_info(name, value):
    # type: (str, str) -> None

    check_type_string("name", name)
    check_type_string("value", value)

    report = get_report()
    report.add_info(name, value)


def is_location_successful(location):
    return get_session().is_successful(location)


def is_session_successful():
    return get_session().is_successful()


def start_test_session():
    return get_session().start_test_session()


def end_test_session():
    return get_session().end_test_session()


def start_test_session_setup():
    return get_session().start_test_session_setup()


def end_test_session_setup():
    return get_session().end_test_session_setup()


def start_test_session_teardown():
    return get_session().start_test_session_teardown()


def end_test_session_teardown():
    return get_session().end_test_session_teardown()


def start_suite(suite):
    get_session().start_suite(suite)


def end_suite(suite):
    get_session().end_suite(suite)


def start_suite_setup(suite):
    get_session().start_suite_setup(suite)


def end_suite_setup(suite):
    get_session().end_suite_setup(suite)


def start_suite_teardown(suite):
    get_session().start_suite_teardown(suite)


def end_suite_teardown(suite):
    get_session().end_suite_teardown(suite)


def start_test(test):
    get_session().start_test(test)


def end_test(test):
    get_session().end_test(test)


def skip_test(test, reason=None):
    get_session().skip_test(test, reason)


def disable_test(test, reason=None):
    get_session().disable_test(test, reason)


class Thread(threading.Thread):
    """
    Acts exactly as the standard threading.Thread class and must be used instead when running threads
    within a test.
    """
    def __init__(self, *args, **kwargs):
        super(Thread, self).__init__(*args, **kwargs)
        self.cursor = get_session().prepare_cursor_for_new_thread()

    def run(self):
        get_session().cursor = self.cursor
        return super(Thread, self).run()



import os.path as osp
import inspect

from typing import Sequence, Any, List
import six

from lemoncheesecake.helpers.moduleimport import get_matching_files, get_py_files_from_dir, strip_py_ext, import_module
from lemoncheesecake.helpers.introspection import get_object_attributes
from lemoncheesecake.exceptions import UserError, ProgrammingError, ModuleImportError, InvalidMetadataError, \
    InvalidSuiteError, VisibilityConditionNotMet, serialize_current_exception
from lemoncheesecake.suite.core import Test, Suite, SUITE_HOOKS
from lemoncheesecake.testtree import BaseTreeNode
from lemoncheesecake.helpers.typecheck import check_type_string, check_type_dict, check_type


def _is_suite_class(obj):
    return inspect.isclass(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_suite


def _is_test_method(obj):
    return inspect.ismethod(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test


def _is_test_function(obj):
    return inspect.isfunction(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test


def _ensure_node_is_visible(obj, metadata):
    if metadata.condition is not None and not metadata.condition(obj):
        raise VisibilityConditionNotMet()


def _check_tags(tags):
    return isinstance(tags, (list, tuple)) and \
           all(isinstance(tag, six.string_types) for tag in tags)


def _check_properties(props):
    return isinstance(props, dict) and \
        all(isinstance(key, six.string_types) for key in props.keys()) and \
        all(isinstance(value, six.string_types) for value in props.values())


def _check_links(links):
    return isinstance(links, (list, tuple)) and \
        all(isinstance(url, six.string_types) for url, _ in links) and \
        all(name is None or isinstance(name, six.string_types) for _, name in links)


def _check_test_tree_node_types(node):
    # type: (BaseTreeNode) -> None
    check_type_string("name", node.name)
    check_type_string("description", node.description)
    check_type("tags", node.tags, "List[str]", _check_tags, show_actual_type=False)
    check_type("properties", node.properties, "Dict[str, str]", _check_properties, show_actual_type=False)
    check_type("links", node.links, "List[Tuple[str, Optional[str]]]", _check_links, show_actual_type=False)


def _load_test(obj):
    md = obj._lccmetadata
    _ensure_node_is_visible(obj, md)

    test = Test(md.name, md.description, obj)
    test.tags.extend(md.tags)
    test.properties.update(md.properties)
    test.links.extend(md.links)
    test.disabled = md.disabled
    test.rank = md.rank
    test.dependencies.extend(md.dependencies)

    try:
        _check_test_tree_node_types(test)
    except TypeError as excp:
        raise InvalidMetadataError("Invalid test metadata for '%s': %s" % (test.name, excp))

    return test


def _load_parametrized_tests(obj):
    md = obj._lccmetadata
    test = _load_test(obj)

    for idx, parameters in enumerate(md.parametrized.parameters_source):
        check_type_dict("test parameters", parameters)
        parametrized_test_name, parametrized_test_description = md.parametrized.naming_scheme(
            test.name, test.description, parameters, idx+1
        )
        parametrized_test = test.pull_node()
        parametrized_test.name = parametrized_test_name
        parametrized_test.description = parametrized_test_description
        parametrized_test.parameters = parameters

        yield parametrized_test


def _load_tests(objs):
    for obj in objs:
        try:
            if obj._lccmetadata.parametrized:
                for test in _load_parametrized_tests(obj):
                    yield test
            else:
                yield _load_test(obj)
        except VisibilityConditionNotMet:
            pass


def _get_test_symbols(obj, filter_func):
    return sorted(
        filter(
            filter_func, (attr for _, attr in get_object_attributes(obj))
        ),
        key=lambda sym: sym._lccmetadata.rank
    )


def _get_test_methods_from_class(obj):
    return _get_test_symbols(obj, _is_test_method)


def _get_sub_suites_from_class(obj):
    return _get_test_symbols(obj, _is_suite_class)


def _get_test_functions_from_module(mod):
    return _get_test_symbols(mod, _is_test_function)


def _get_suite_classes_from_module(mod):
    return _get_test_symbols(mod, _is_suite_class)


def _get_generated_tests(obj):
    return getattr(obj, "_lccgeneratedtests", [])


def load_suite_from_class(class_):
    # type: (Any) -> Suite
    """
    Load a suite from a class.
    """
    try:
        md = class_._lccmetadata
    except AttributeError:
        raise InvalidSuiteError()

    try:
        suite_obj = class_()
    except UserError as e:
        raise e  # propagate UserError
    except Exception:
        raise ProgrammingError("Got an unexpected error while instantiating suite class '%s':%s" % (
            class_.__name__, serialize_current_exception()
        ))
    _ensure_node_is_visible(suite_obj, md)

    suite = Suite(suite_obj, md.name, md.description)
    suite.tags.extend(md.tags)
    suite.properties.update(md.properties)
    suite.links.extend(md.links)
    suite.rank = md.rank
    suite.disabled = md.disabled

    try:
        _check_test_tree_node_types(suite)
    except TypeError as excp:
        raise InvalidMetadataError("Invalid suite metadata for '%s': %s" % (suite.name, excp))

    for hook_name in SUITE_HOOKS:
        if hasattr(suite_obj, hook_name):
            suite.add_hook(hook_name, getattr(suite_obj, hook_name))

    for test in _load_tests(_get_test_methods_from_class(suite_obj)):
        suite.add_test(test)

    for test in _get_generated_tests(suite_obj):
        suite.add_test(test)

    for sub_suite in load_suites_from_classes(_get_sub_suites_from_class(suite_obj)):
        suite.add_suite(sub_suite)

    return suite


def load_suites_from_classes(classes):
    # type: (Sequence[Any]) -> List[Suite]
    """
    Load a list of suites from a list of classes.
    """
    suites = []
    for class_ in classes:
        try:
            suites.append(load_suite_from_class(class_))
        except VisibilityConditionNotMet:
            pass
    return suites


def load_suite_from_module(mod):
    # type: (Any) -> Suite
    """
    Load a suite from a module object.
    """
    # TODO: find a better way to workaround circular import
    from lemoncheesecake.suite.builder import _get_metadata_next_rank

    suite_info = getattr(mod, "SUITE")
    suite_condition = suite_info.get("visible_if")
    if suite_condition is not None and not suite_condition(mod):
        raise VisibilityConditionNotMet()

    suite_name = inspect.getmodulename(inspect.getfile(mod))

    try:
        suite_description = suite_info["description"]
    except KeyError:
        raise InvalidMetadataError("Missing description in '%s' suite information" % mod.__file__)

    suite = Suite(mod, suite_name, suite_description)
    suite.tags.extend(suite_info.get("tags", []))
    suite.properties.update(suite_info.get("properties", []))
    suite.links.extend(suite_info.get("links", []))
    suite.rank = suite_info.get("rank", _get_metadata_next_rank())

    try:
        _check_test_tree_node_types(suite)
    except TypeError as excp:
        raise InvalidMetadataError("Invalid suite metadata for '%s': %s" % (suite.name, excp))

    for hook_name in SUITE_HOOKS:
        if hasattr(mod, hook_name):
            suite.add_hook(hook_name, getattr(mod, hook_name))

    for test in _load_tests(_get_test_functions_from_module(mod)):
        suite.add_test(test)

    for test in _get_generated_tests(mod):
        suite.add_test(test)

    for sub_suite in load_suites_from_classes(_get_suite_classes_from_module(mod)):
        suite.add_suite(sub_suite)

    return suite


def load_suite_from_file(filename):
    # type: (str) -> Suite

    """
    Load a suite from a Python module indicated by a filename.

    A valid module is either:
      - a module containing a dict name 'SUITE' with keys:
            - description (mandatory)
            - tags (optional)
            - properties (optional)
            - links (optional)
            - rank (optional)

            Example::

                SUITE = {
                    "description": "Such a great test suite"
                }
                @lcc.test("Such a great test")
                def my_test():
                    pass

      - a module that contains a suite class with the same name as the module name

            Example::

                @lcc.suite("Such a great test suite")
                class my_suite:
                    @lcc.test("Such a great test")
                    def my_test():
                        pass

    Raise ModuleImportError if the suite class cannot be imported.
    """
    mod = import_module(filename)

    if hasattr(mod, "SUITE"):
        suite = load_suite_from_module(mod)
    else:
        mod_name = strip_py_ext(osp.basename(filename))
        try:
            klass = getattr(mod, mod_name)
        except AttributeError:
            raise ModuleImportError("Cannot find class '%s' in '%s'" % (mod_name, mod.__file__))

        suite = load_suite_from_class(klass)
    return suite


def load_suites_from_files(patterns, excluding=()):
    # type: (Sequence[str], Sequence[str]) -> List[Suite]

    """
    Load a list of suites from a list of files.

    :param patterns: a mandatory list (a simple string can also be used instead of a single element list)
      of files to import; the wildcard '*' character can be used
    :param exclude: an optional list (a simple string can also be used instead of a single element list)
      of elements to exclude from the expanded list of files to import

    Example::

        load_suites_from_files("test_*.py")
    """
    suites = []
    for filename in get_matching_files(patterns, excluding):
        try:
            suites.append(load_suite_from_file(filename))
        except VisibilityConditionNotMet:
            pass
    return suites


def load_suites_from_directory(dir, recursive=True):
    # type: (str, bool) -> List[Suite]

    """
    Load a list of suites from a directory.

    The function expect that:
      - each module (.py file) contains a class that inherits Suite
      - the class name must have the same name as the module name (if the module is foo.py
        the class must be named foo)

    If the recursive argument is set to True, sub suites will be searched in a directory named
    from the suite module: if the suite module is "foo.py" then the sub suites directory must be "foo".

    Raise ModuleImportError if one or more suite cannot be imported.
    """
    if not osp.exists(dir):
        raise ModuleImportError("Directory '%s' does not exist" % dir)

    suites = []
    for filename in get_py_files_from_dir(dir):
        try:
            suite = load_suite_from_file(filename)
        except VisibilityConditionNotMet:
            continue

        if recursive:
            subsuites_dir = strip_py_ext(filename)
            if osp.isdir(subsuites_dir):
                for sub_suite in load_suites_from_directory(subsuites_dir, recursive=True):
                    suite.add_suite(sub_suite)
        suites.append(suite)
    suites.sort(key=lambda suite: suite.rank)
    return suites



'''
Created on Mar 27, 2017

@author: nicolas
'''

from typing import Any, Mapping, List
import six

from lemoncheesecake.session import log_check
from lemoncheesecake.exceptions import AbortTest
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.dict_ import HasEntry, wrap_key_matcher
from lemoncheesecake.matching.matchers.composites import is_


class _HasEntry(HasEntry):
    def build_description(self, _):
        ret = self.key_matcher.build_description()
        if self.value_matcher:
            ret += " " + self.value_matcher.build_description(MatcherDescriptionTransformer())
        return ret


def _build_matcher_for_dict_operation(key_matcher, value_matcher, base_key):
    return _HasEntry(
        wrap_key_matcher(key_matcher, base_key=base_key),
        value_matcher if value_matcher is not None else is_(value_matcher)
    )


def _format_result_details(details):
    if details is None:
        return None

    if not isinstance(details, six.string_types):
        details = str(details)

    return details[0].upper() + details[1:]


def _log_match_result(hint, matcher, result, quiet=False):
    if hint is not None:
        description = "Expect %s %s" % (hint, matcher.build_description(MatcherDescriptionTransformer()))
    else:
        description = "Expect %s" % matcher.build_description(MatcherDescriptionTransformer())

    return log_check(
        description, result.is_successful, _format_result_details(result.description) if not quiet else None
    )


def check_that(hint, actual, matcher, quiet=False):
    # type: (str, Any, Matcher, bool) -> MatchResult

    """Check that actual matches given matcher.

    A check log is added to the report.

    If ``quiet`` is set to ``True``, the check details won't appear in the check log.
    """
    assert isinstance(matcher, Matcher)

    result = matcher.matches(actual)
    _log_match_result(hint, matcher, result, quiet=quiet)
    return result


def _do_that_in(func, actual, *args, **kwargs):
    if len(args) % 2 != 0:
        raise TypeError("function expects an even number of *args")

    for value_matcher in args[1::2]:  # iterate over each odd element
        assert value_matcher is None or isinstance(value_matcher, Matcher)

    base_key = kwargs.get("base_key", [])
    quiet = kwargs.get("quiet", False)

    results = []
    i = 0
    while i < len(args):
        key_matcher, value_matcher = args[i], args[i+1]
        result = func(
            None, actual, _build_matcher_for_dict_operation(key_matcher, value_matcher, base_key), quiet=quiet
        )
        results.append(result)
        i += 2

    return results


def check_that_in(actual, *args, **kwargs):
    # type: (Mapping, Any, Any) -> List[MatchResult]

    """
    Equivalent of :py:func:`check_that` over items of a dict.

    Example of usage::

        check_that_in(
            {"foo": 1, "bar": 2},
            "foo", equal_to(1),
            "bar", equal_to(2)
        )

    The key can also be a ``tuple`` when checking for a nested item::

        check_that_in(
            {"foo": {"bar": 2}},
            ("foo", "bar"), equal_to(2)
        )

    The function can take a ``base_key`` keyword-arg to pass repeating nested-keys as a ``tuple``.

    If an extra ``quiet`` keyword-arg is set to ``True``, the check details won't appear in the check log.

    The function returns a list of :py:class:`MatchResult <lemoncheesecake.matching.matcher.MatchResult>`.
    """

    return _do_that_in(check_that, actual, *args, **kwargs)


def require_that_in(actual, *args, **kwargs):
    # type: (Mapping, Any, Any) -> List[MatchResult]

    """
    Does the same thing as :py:func:`check_that_in` except it performs a
    :py:func:`require_that` on each key-value pair.
    """

    return _do_that_in(require_that, actual, *args, **kwargs)


def assert_that_in(actual, *args, **kwargs):
    # type: (Mapping, Any, Any) -> List[MatchResult]

    """
    Does the same thing as :py:func:`check_that_in` except it performs a
    :py:func:`assert_that` on each key-value pair.
    """

    return _do_that_in(assert_that, actual, *args, **kwargs)


def require_that(hint, actual, matcher, quiet=False):
    # type: (str, Any, Matcher, bool) -> MatchResult

    """Require that actual matches given matcher.

    A check log is added to the report. An AbortTest exception is raised if the check does not succeed.

    If ``quiet`` is set to ``True``, the check details won't appear in the check log.
    """
    assert isinstance(matcher, Matcher)

    result = matcher.matches(actual)
    _log_match_result(hint, matcher, result, quiet=quiet)
    if result:
        return result
    else:
        raise AbortTest("previous requirement was not fulfilled")


def assert_that(hint, actual, matcher, quiet=False):
    # type: (str, Any, Matcher, bool) -> MatchResult

    """Assert that actual matches given matcher.

    If assertion fail, a check log is added to the report and an AbortTest exception is raised.

    If ``quiet`` is set to ``True``, the check details won't appear in the check log.
    """
    assert isinstance(matcher, Matcher)

    result = matcher.matches(actual)
    if result:
        return result
    else:
        _log_match_result(hint, matcher, result, quiet=quiet)
        raise AbortTest("assertion error")



'''
Created on Mar 26, 2016

@author: nicolas
'''

import time
from decimal import Decimal
from functools import reduce
from typing import Union, List, Iterable, Optional
import datetime
import calendar

from typing import Callable, Any

from lemoncheesecake.consts import LOG_LEVEL_ERROR
from lemoncheesecake.helpers.time import humanize_duration
from lemoncheesecake.testtree import BaseTest, BaseSuite, flatten_tests, flatten_suites, find_test, find_suite, \
    filter_suites, normalize_node_hierarchy, TreeNodeHierarchy

_TEST_STATUSES = "passed", "failed", "skipped", "disabled"
_DEFAULT_REPORT_TITLE = "Test Report"


# NB: it would be nicer to use:
# datetime.isoformat(sep=' ', timespec='milliseconds')
# unfortunately, the timespec argument is only available since Python 3.6
def format_time_as_iso8601(ts):
    # type: (float) -> str
    """
    Serialize time as ISO8601, such as: "2019-05-04T22:57:08.399Z"
    """
    ts = round(ts, 3)  # round the timestamp to milliseconds
    result = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(ts))  # builds datetime with second precision
    result += ".%03d" % (Decimal(repr(ts)) % 1 * 1000)  # adds millisecond precision
    result += "Z"  # declares as UTC
    return result


def parse_iso8601_time(s):
    # type: (str) -> float
    """
    Parse time as generated by format_time_as_iso8601
    """
    dt = datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ")
    return calendar.timegm(dt.utctimetuple()) + float(dt.microsecond) / 1000000


def _get_duration(start_time, end_time):
    # type: (Optional[float], Optional[float]) -> Optional[float]
    if start_time is not None and end_time is not None:
        return end_time - start_time
    else:
        return None


class Log(object):
    def __init__(self, level, message, ts):
        # type: (str, str, float) -> None
        self.level = level
        self.message = message
        self.time = ts


class Check(object):
    def __init__(self, description, is_successful, details, ts):
        # type: (str, bool, Optional[str], float) -> None
        self.description = description
        self.is_successful = is_successful
        self.details = details
        self.time = ts


class Attachment(object):
    def __init__(self, description, filename, as_image, ts):
        # type: (str, str, bool, float) -> None
        self.description = description
        self.filename = filename
        self.as_image = as_image
        self.time = ts


class Url(object):
    def __init__(self, description, url, ts):
        # type: (str, str, float) -> None
        self.description = description
        self.url = url
        self.time = ts


class Step(object):
    def __init__(self, description, detached=False):
        # type: (str, bool) -> None
        self.description = description
        self._detached = detached  # this attribute is runtime only is not intended to be serialized
        self.parent_result = None
        self.entries = []  # type: List[Union[Log, Check, Attachment, Url]]
        self.start_time = None  # type: Optional[float]
        self.end_time = None  # type: Optional[float]

    @staticmethod
    def _is_entry_successful(entry):
        if isinstance(entry, Check):
            return entry.is_successful
        elif isinstance(entry, Log):
            return entry.level != LOG_LEVEL_ERROR
        return True

    def is_successful(self):
        # type: () -> bool
        return all(map(Step._is_entry_successful, self.entries))

    @property
    def duration(self):
        # type: () -> Optional[float]
        return _get_duration(self.start_time, self.end_time)


class Result(object):
    def __init__(self):
        # please note that this attribute is also defined in TestResult through BaseTest,
        # one will override the other
        self.parent_suite = None  # type: Optional[SuiteResult]
        self.type = None  # type: Optional[str]
        self._steps = []  # type: List[Step]
        self.start_time = None  # type: Optional[float]
        self.end_time = None  # type: Optional[float]
        self.status = None  # type: Optional[str]
        self.status_details = None  # type: Optional[str]

    def add_step(self, step):
        # type: (Step) -> None
        step.parent_result = self
        self._steps.append(step)

    def get_steps(self):
        # type: () -> List[Step]
        return self._steps

    def is_successful(self):
        # type: () -> bool
        if self.status:  # test is finished
            return self.status in ("passed", "disabled")
        else:  # check if the test is successful "so far"
            return all(step.is_successful() for step in self._steps)

    @property
    def duration(self):
        # type: () -> Optional[float]
        return _get_duration(self.start_time, self.end_time)


class TestResult(BaseTest, Result):
    def __init__(self, name, description):
        BaseTest.__init__(self, name, description)
        Result.__init__(self)
        # non-serialized attributes (only set in-memory during test execution):
        self.rank = 0


class SuiteResult(BaseSuite):
    def __init__(self, name, description):
        BaseSuite.__init__(self, name, description)
        self.start_time = None  # type: Optional[float]
        self.end_time = None  # type: Optional[float]
        self._suite_setup = None  # type: Optional[Result]
        self._suite_teardown = None  # type: Optional[Result]
        # non-serialized attributes (only set in-memory during test execution)
        self.rank = 0

    @property
    def suite_setup(self):
        return self._suite_setup
    
    @suite_setup.setter
    def suite_setup(self, setup):
        if setup:
            setup.parent_suite = self
            setup.type = "suite_setup"
        self._suite_setup = setup

    @property
    def suite_teardown(self):
        return self._suite_teardown

    @suite_teardown.setter
    def suite_teardown(self, teardown):
        if teardown:
            teardown.parent_suite = self
            teardown.type = "suite_teardown"
        self._suite_teardown = teardown

    @property
    def duration(self):
        # type: () -> Optional[float]
        return reduce(
            lambda x, y: x + y,
            # result.duration is None if the corresponding result is in progress
            [result.duration or 0 for result in flatten_results([self])],
            0
        )

    def get_tests(self):
        # type: () -> List[TestResult]
        tests = super(SuiteResult, self).get_tests()
        return sorted(tests, key=lambda t: t.rank)

    def get_suites(self, include_empty_suites=False):
        # type: (bool) -> List[SuiteResult]
        suites = super(SuiteResult, self).get_suites(include_empty_suites)
        return sorted(suites, key=lambda s: s.rank)

    def pull_node(self):
        node = BaseSuite.pull_node(self)
        node._suite_setup = None
        node._suite_teardown = None
        return node

    def is_empty(self):
        return BaseSuite.is_empty(self) and self._suite_setup is None and self._suite_teardown is None

    def filter(self, result_filter):
        suite = BaseSuite.filter(self, result_filter)
        if self._suite_setup and result_filter(self._suite_setup):
            suite._suite_setup = self._suite_setup
        if self._suite_teardown and result_filter(self._suite_teardown):
            suite._suite_teardown = self._suite_teardown
        return suite


class ReportLocation(object):
    _TEST_SESSION_SETUP = 0
    _TEST_SESSION_TEARDOWN = 1
    _SUITE_SETUP = 2
    _SUITE_TEARDOWN = 3
    _TEST = 4

    def __init__(self, node_type, node_hierarchy=None):
        self.node_type = node_type
        self.node_hierarchy = node_hierarchy

    @classmethod
    def in_test_session_setup(cls):
        # type: () -> ReportLocation
        return cls(cls._TEST_SESSION_SETUP)

    @classmethod
    def in_test_session_teardown(cls):
        # type: () -> ReportLocation
        return cls(cls._TEST_SESSION_TEARDOWN)

    @classmethod
    def in_suite_setup(cls, suite):
        # type: (TreeNodeHierarchy) -> ReportLocation
        return cls(cls._SUITE_SETUP, normalize_node_hierarchy(suite))

    @classmethod
    def in_suite_teardown(cls, suite):
        # type: (TreeNodeHierarchy) -> ReportLocation
        return cls(cls._SUITE_TEARDOWN, normalize_node_hierarchy(suite))

    @classmethod
    def in_test(cls, test):
        # type: (TreeNodeHierarchy) -> ReportLocation
        return cls(cls._TEST, normalize_node_hierarchy(test))

    def get(self, report):
        # type: (Report) -> Union[Result, SuiteResult, TestResult, None]
        if self.node_type == self._TEST_SESSION_SETUP:
            return report.test_session_setup
        elif self.node_type == self._TEST_SESSION_TEARDOWN:
            return report.test_session_teardown
        elif self.node_type == self._SUITE_SETUP:
            suite = report.get_suite(self.node_hierarchy)
            return suite.suite_setup
        elif self.node_type == self._SUITE_TEARDOWN:
            suite = report.get_suite(self.node_hierarchy)
            return suite.suite_teardown
        elif self.node_type == self._TEST:
            return report.get_test(self.node_hierarchy)
        else:
            raise Exception("Unknown self type %s" % self.node_type)

    def __eq__(self, other):
        return all((
            isinstance(other, ReportLocation),
            self.node_type == other.node_type,
            self.node_hierarchy == other.node_hierarchy
        ))

    def __hash__(self):
        return hash((self.node_type, self.node_hierarchy))

    def __str__(self):
        ret = ""
        if self.node_hierarchy:
            ret += ".".join(self.node_hierarchy) + " "
        ret += ("session setup", "session teardown", "suite setup", "suite teardown", "test")[self.node_type]
        return "<%s>" % ret


def flatten_results(suites):
    # type: (Iterable[SuiteResult]) -> Iterable[Result]
    for suite in flatten_suites(suites):
        if suite.suite_setup:
            yield suite.suite_setup
        for test in suite.get_tests():
            yield test
        if suite.suite_teardown:
            yield suite.suite_teardown


class Report(object):
    def __init__(self):
        self.info = []
        self._test_session_setup = None  # type: Optional[Result]
        self._test_session_teardown = None  # type: Optional[Result]
        self._suites = []  # type: List[SuiteResult]
        self.start_time = None  # type: Optional[float]
        self.end_time = None  # type: Optional[float]
        self.report_generation_time = None  # type: Optional[float]
        self.title = _DEFAULT_REPORT_TITLE
        self.nb_threads = 1

    @property
    def test_session_setup(self):
        return self._test_session_setup

    @test_session_setup.setter
    def test_session_setup(self, setup):
        if setup:
            setup.type = "test_session_setup"
        self._test_session_setup = setup

    @property
    def test_session_teardown(self):
        return self._test_session_teardown

    @test_session_teardown.setter
    def test_session_teardown(self, teardown):
        if teardown:
            teardown.type = "test_session_teardown"
        self._test_session_teardown = teardown

    @property
    def duration(self):
        # type: () -> Optional[float]
        return _get_duration(self.start_time, self.end_time)

    @property
    def nb_tests(self):
        # type: () -> int
        return len(list(self.all_tests()))

    @property
    def parallelized(self):
        # type: () -> bool
        return self.nb_threads > 1 and self.nb_tests > 1

    def add_info(self, name, value):
        # type: (str, str) -> None
        self.info.append([name, value])
    
    def add_suite(self, suite):
        # type: (SuiteResult) -> None
        self._suites.append(suite)
    
    def get_suites(self):
        # type: () -> List[SuiteResult]
        return sorted(self._suites, key=lambda s: s.rank)

    def get_suite(self, hierarchy):
        # type: (List[str]) -> SuiteResult
        return find_suite(self._suites, hierarchy)

    def get_test(self, hierarchy):
        # type: (List[str]) -> TestResult
        return find_test(self._suites, hierarchy)

    def get(self, location):
        # type: (ReportLocation) -> Union[Result, SuiteResult, TestResult, None]
        return location.get(self)

    def is_successful(self):
        # type: () -> bool
        return all(test.status in ("passed", "disabled") for test in self.all_tests())

    def all_suites(self, result_filter=None):
        # type: (Optional[Callable[[TestResult], bool]]) -> Iterable[SuiteResult]
        if result_filter:
            return flatten_suites(filter_suites(self._suites, result_filter))
        else:
            return flatten_suites(self._suites)

    def all_tests(self):
        # type: () -> Iterable[TestResult]
        return flatten_tests(self._suites)

    def all_results(self):
        # type: () -> Iterable[Result]
        if self.test_session_setup:
            yield self.test_session_setup
        for result in flatten_results(self.get_suites()):
            yield result
        if self.test_session_teardown:
            yield self.test_session_teardown

    def all_steps(self):
        # type: () -> Iterable[Step]
        for result in self.all_results():
            for step in result.get_steps():
                yield step


class ReportStats(object):
    def __init__(self):
        self.tests_nb = 0
        self.tests_nb_by_status = {s: 0 for s in _TEST_STATUSES}
        self.duration = None
        self.duration_cumulative = 0

    @property
    def tests_enabled_nb(self):
        return sum((self.tests_nb_by_status["passed"], self.tests_nb_by_status["failed"], self.tests_nb_by_status["skipped"]))

    @property
    def successful_tests_percentage(self):
        return (float(self.tests_nb_by_status["passed"]) / self.tests_enabled_nb * 100) if self.tests_enabled_nb else 0

    @property
    def duration_cumulative_description(self):
        description = humanize_duration(self.duration_cumulative)
        if self.duration:
            description += " (parallelization speedup factor is %.1f)" % (float(self.duration_cumulative) / self.duration)
        return description

    @classmethod
    def from_results(cls, results, duration):
        # type: (List[Result], Any[int, None]) -> ReportStats

        stats = cls()

        stats.duration = duration

        stats.duration_cumulative = sum(result.duration or 0 for result in results)

        tests = list(filter(lambda r: isinstance(r, TestResult), results))

        stats.tests_nb = len(tests)

        for test in tests:
            if test.status:
                stats.tests_nb_by_status[test.status] += 1

        return stats

    @classmethod
    def from_report(cls, report):
        # type: (Report) -> ReportStats
        return cls.from_results(list(report.all_results()), report.duration)

    @classmethod
    def from_suites(cls, suites, parallelized):
        # type: (List[SuiteResult], bool) -> ReportStats
        results = list(flatten_results(suites))

        return cls.from_results(
            results,
            results[-1].end_time - results[0].start_time if not parallelized else None
        )



from __future__ import print_function

import os
import sys
import time

try:
    import slacker
    SLACKER_IS_AVAILABLE = True
except ImportError:
    SLACKER_IS_AVAILABLE = False

from lemoncheesecake.reporting import ReportingBackend, ReportingSession, ReportingSessionBuilderMixin, ReportStats
from lemoncheesecake.exceptions import UserError
from lemoncheesecake.helpers.time import humanize_duration


def percent(val, of):
    return "%d%%" % ((float(val) / of * 100) if of else 0)


def get_message_template_parameters():
    return {
        "start_time": lambda report, stats: time.asctime(time.localtime(report.start_time)),
        "end_time": lambda report, stats: time.asctime(time.localtime(report.end_time)),
        "duration": lambda report, stats: humanize_duration(report.end_time - report.start_time),

        "total": lambda report, stats: stats.tests_nb,
        "enabled": lambda report, stats: stats.tests_enabled_nb,

        "passed": lambda report, stats: stats.tests_nb_by_status["passed"],
        "passed_pct": lambda report, stats: percent(stats.tests_nb_by_status["passed"], of=stats.tests_enabled_nb),

        "failed": lambda report, stats: stats.tests_nb_by_status["failed"],
        "failed_pct": lambda report, stats: percent(stats.tests_nb_by_status["failed"], of=stats.tests_enabled_nb),

        "skipped": lambda report, stats: stats.tests_nb_by_status["skipped"],
        "skipped_pct": lambda report, stats: percent(stats.tests_nb_by_status["skipped"], of=stats.tests_enabled_nb),

        "disabled": lambda report, stats: stats.tests_nb_by_status["disabled"],
        "disabled_pct": lambda report, stats: percent(stats.tests_nb_by_status["disabled"], of=stats.tests_nb)
    }


def build_message_parameters(report):
    stats = ReportStats.from_report(report)
    return {
        name: func(report, stats) for name, func in get_message_template_parameters().items()
    }


def validate_message_template_validity(template):
    try:
        template.format(**{name: "" for name in get_message_template_parameters()})
        return template
    except KeyError as excp:
        raise UserError("Invalid Slack message template, unknown variable: %s" % excp)


class SlackReportingSession(ReportingSession):
    def __init__(self, auth_token, channel, message_template, proxy=None, only_notify_failure=False):
        self.slacker = slacker.Slacker(auth_token, http_proxy=proxy, https_proxy=proxy)
        self.channel = channel
        self.message_template = message_template
        self.only_notify_failure = only_notify_failure

    def on_test_session_end(self, event):
        if self.only_notify_failure and event.report.is_successful():
            return

        message = self.message_template.format(**build_message_parameters(event.report))
        try:
            self.slacker.chat.post_message(self.channel, message)
        except slacker.Error as excp:
            print("Error while notifying Slack channel/user '%s', got: %s" % (
                self.channel, excp
            ), file=sys.stderr)


def get_env_var(name, optional=False, default=None):
    try:
        return os.environ[name]
    except KeyError as excp:
        if optional:
            return default
        else:
            raise UserError("Missing environment variable '%s' for Slack reporting backend" % excp)


class SlackReportingBackend(ReportingBackend, ReportingSessionBuilderMixin):
    def get_name(self):
        return "slack"

    def is_available(self):
        return SLACKER_IS_AVAILABLE

    def create_reporting_session(self, report_dir, report, parallel, saving_strategy):
        return SlackReportingSession(
            auth_token=get_env_var("LCC_SLACK_AUTH_TOKEN"),
            channel=get_env_var("LCC_SLACK_CHANNEL"),
            message_template=validate_message_template_validity(get_env_var(
                "LCC_SLACK_MESSAGE_TEMPLATE", optional=True,
                default="{passed} passed, {failed} failed, {skipped} skipped"
            )),
            proxy=get_env_var("LCC_SLACK_HTTP_PROXY", optional=True, default=None),
            only_notify_failure="LCC_SLACK_ONLY_NOTIFY_FAILURE" in os.environ
        )



'''
Created on Oct 31, 2016

@author: nicolas
'''

ATTACHMENTS_DIR = "attachments"

LOG_LEVEL_DEBUG = "debug"
LOG_LEVEL_INFO = "info"
LOG_LEVEL_WARN = "warn"
LOG_LEVEL_ERROR = "error"

NEGATIVE_CHARS = "-^~"



from typing import Optional

import sys
import traceback


class LemoncheesecakeException(Exception):
    message_prefix = None  # type: Optional[str]

    def __str__(self):
        s = Exception.__str__(self)
        if self.message_prefix:
            s = "%s, %s" % (self.message_prefix, s)
        return s


class ProgrammingError(LemoncheesecakeException):
    pass


class ProjectError(LemoncheesecakeException):
    pass


class ModuleImportError(LemoncheesecakeException):
    pass


class FixtureError(LemoncheesecakeException):
    pass


class InvalidSuiteError(LemoncheesecakeException):
    pass


class InvalidMetadataError(ProgrammingError):
    pass


class VisibilityConditionNotMet(LemoncheesecakeException):
    pass


class UnknownReportBackendError(LemoncheesecakeException):
    pass


class AbortTest(LemoncheesecakeException):
    """
    Raising this exception will stop the currently running test.
    """

    message_prefix = "The test has been aborted"

    def __init__(self, reason):
        LemoncheesecakeException.__init__(self, reason)


class AbortSuite(LemoncheesecakeException):
    """
    Raising this exception will stop the currently running suite.
    """
    message_prefix = "The suite has been aborted"

    def __init__(self, reason):
        LemoncheesecakeException.__init__(self, reason)


class AbortAllTests(LemoncheesecakeException):
    """
    Raising this exception will stop the currently running test and all the tests waiting to be run.
    """
    message_prefix = "All tests have been aborted"

    def __init__(self, reason):
        LemoncheesecakeException.__init__(self, reason)


class UserError(LemoncheesecakeException):
    """
    This exception is intended to be raised in pre-run and post-run phases of the project
    to indicate that a required state has not been fulfilled.
    """
    # NB: sphinx requires the constructor to be overriden, otherwise it raises an error
    def __init__(self, reason):
        LemoncheesecakeException.__init__(self, reason)


class InvalidReportFile(LemoncheesecakeException):
    pass


class IncompatibleReportFile(LemoncheesecakeException):
    pass


class CannotFindTreeNode(LemoncheesecakeException):
    pass


class TaskFailure(LemoncheesecakeException):
    pass


class TasksExecutionFailure(LemoncheesecakeException):
    pass


class CircularDependencyError(LemoncheesecakeException):
    pass


def serialize_current_exception(show_stacktrace=True):
    if show_stacktrace:
        return "\n" + "<" * 72 + "\n" + traceback.format_exc() + ">" * 72
    else:
        return " " + str(sys.exc_info()[1])



'''
Created on Sep 8, 2016

@author: nicolas
'''

import inspect
import copy

from typing import Any, Iterable, Callable, Optional, Tuple

from lemoncheesecake.suite.core import InjectedFixture, Test
from lemoncheesecake.exceptions import ProgrammingError


class Metadata(object):
    _next_rank = 1

    def __init__(self):
        self.is_test = False
        self.is_suite = False
        self.name = None
        self.description = None
        self.properties = {}
        self.tags = []
        self.links = []
        self.rank = 0
        self.dependencies = []
        self.disabled = False
        self.condition = None
        self.parametrized = None


def _get_metadata_next_rank():
    rank = Metadata._next_rank
    Metadata._next_rank += 1
    return rank


def add_test_into_suite(test, suite):
    # type: (Test, Any) -> None
    """
    Add test into suite

    :param test: a :py:class:`Test <lemoncheesecake.suite.core.Test>` instance
    :param suite: a suite decorated class instance (in that case the function must be
        called when the class is instantiated) or a module marked as a suite (in that case the function
        must be called when the module is loaded)
    """
    if not hasattr(suite, "_lccgeneratedtests"):
        suite._lccgeneratedtests = []
    if test.rank is None:
        test.rank = _get_metadata_next_rank()
    suite._lccgeneratedtests.append(test)


_objects_with_metadata = []
def get_metadata(obj):
    global _objects_with_metadata

    if hasattr(obj, "_lccmetadata"):
        if obj not in _objects_with_metadata:  # metadata comes from the superclass
            obj._lccmetadata = copy.deepcopy(obj._lccmetadata)
        return obj._lccmetadata
    else:
        obj._lccmetadata = Metadata()
        _objects_with_metadata.append(obj)
        return obj._lccmetadata


def suite(description, name=None, rank=None):
    """
    Decorator, mark a class as a suite class.

    :param description: suite's description
    :param name: suite's name (by default, the suite's name is taken from the class's name)
    :param rank: this value is used to order suites of the same hierarchy level
    """
    def wrapper(klass):
        if not inspect.isclass(klass):
            raise ProgrammingError("%s is not a class (suite decorator can only be used on a class)" % klass)
        md = get_metadata(klass)
        if md.dependencies:
            raise ProgrammingError("'depends_on' can not be used on a suite class")
        md.is_suite = True
        md.rank = rank if rank is not None else _get_metadata_next_rank()
        md.name = name or klass.__name__
        md.description = description
        return klass
    return wrapper


def test(description, name=None):
    """
    Decorator, make a method as a test method.

    :param description: test's description
    :param name: test's name (by default, the suite's name is taken from the class's name)
    """
    def wrapper(func):
        if not inspect.isfunction(func):
            raise ProgrammingError("%s is not a function (test decorator can only be used on a function)" % func)
        md = get_metadata(func)
        md.is_test = True
        md.rank = _get_metadata_next_rank()
        md.name = name or func.__name__
        md.description = description
        return func
    return wrapper


def tags(*tag_names):
    """Decorator, add tags to a test or a suite."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.tags.extend(tag_names)
        return obj
    return wrapper


def prop(key, value):
    """Decorator, add a property (key/value) to a test or a suite."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.properties[key] = value
        return obj
    return wrapper


def link(url, name=None):
    """Decorator, set a link (with an optional friendly name) to a test or a suite."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.links.append((url, name))
        return obj
    return wrapper


def disabled(reason=None):
    """Decorator, mark a test or a suite as disabled, meaning it won't be executed but will be visible
    in report. An optional reason can be passed to the decorator (*new in version 1.1.0*)."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.disabled = reason if reason else True
        return obj
    return wrapper


def visible_if(condition):
    """
    Decorator, the test or suite will only appear if the given callable return a true value.

    :param condition: a callable that will take the test object if applied to a test or the suite class
        instance if applied to a suite.
    """
    def wrapper(obj):
        md = get_metadata(obj)
        md.condition = condition
        return obj
    return wrapper


def hidden():
    """Decorator, the test or suite won't be visible in the resulting test tree."""
    return visible_if(lambda _: False)


def depends_on(*deps):
    # type: (*str) -> Any
    """
    Decorator, only applicable to a test. Add dependencies to a test.

    :param deps: the test paths that the decorated test is depending on.
    """
    def wrapper(obj):
        md = get_metadata(obj)
        if md.is_suite:
            raise ProgrammingError("'depends_on' can not be used on a suite class")
        md.dependencies.extend(deps)
        return obj
    return wrapper


def inject_fixture(fixture_name=None):
    """
    Inject a fixture into a suite. If no fixture name is specified then the name of the variable holding
    the injected fixture will be used.
    """
    return InjectedFixture(fixture_name)


def _default_naming_scheme(name, description, parameters, nb):
    return name + "_%d" % nb, description + " #%d" % nb


class _Parametrized(object):
    def __init__(self, parameters_source, naming_scheme):
        self.parameters_source = parameters_source
        self.naming_scheme = naming_scheme


def parametrized(parameter_source, naming_scheme=_default_naming_scheme):
    # type: (Iterable[dict], Optional[Callable[[str, str, dict, int], Tuple[str, str]]]) -> Any
    """
    Decorator, make the test parametrized.

    :param parameter_source: a list / iterable of dicts, each dict represent the parameters that will be used
        by the test
    :param naming_scheme: a optional function that takes as parameters the base test name, description, parameters, index
         and must return the expanded test name and description as a two elements list
    """

    def wrapper(obj):
        md = get_metadata(obj)
        md.parametrized = _Parametrized(parameter_source, naming_scheme)
        return obj
    return wrapper



'''
Created on Dec 10, 2016

@author: nicolas
'''

import sys
import os
import os.path as osp
import shutil
import argparse

from typing import Any, Dict, Optional, Sequence, Tuple, List

from lemoncheesecake.suite import load_suites_from_directory, Suite
from lemoncheesecake.fixture import load_fixtures_from_directory, Fixture
from lemoncheesecake.metadatapolicy import MetadataPolicy
from lemoncheesecake.reporting import get_reporting_backends, ReportingBackend
from lemoncheesecake.reporting.reportdir import create_report_dir_with_rotation
from lemoncheesecake.exceptions import ProjectError, UserError, serialize_current_exception
from lemoncheesecake.helpers.resources import get_resource_path
from lemoncheesecake.helpers.moduleimport import import_module

PROJECT_CONFIG_FILE = "project.py"
DEFAULT_REPORTING_BACKENDS = ("console", "json", "html")


class Project(object):
    def __init__(self, project_dir):
        #: The project's directory path
        self.dir = project_dir  # type: str
        #: The project's metadata policy
        self.metadata_policy = MetadataPolicy()
        #: Indicates whether or not the project supports parallel execution of tests
        self.threaded = True  # type: bool
        #: Indicated whether or not the command line ("lcc run...") will be displayed in the report
        self.show_command_line_in_report = True  # type: bool
        #: The reporting backends of the project as a dict (whose key is the reporting backend name)
        self.reporting_backends = {b.get_name(): b for b in get_reporting_backends()}  # type: Dict[str, ReportingBackend]
        #: The list of default reporting backend (indicated by their name) that will be used by "lcc run"
        self.default_reporting_backend_names = list(DEFAULT_REPORTING_BACKENDS)

    def add_cli_args(self, cli_parser):
        # type: (argparse.ArgumentParser) -> None
        """
        Overridable. This method can be used to add extra CLI arguments to "lcc run".
        """
        pass

    def create_report_dir(self):
        # type: () -> str
        """
        Overridable. Create the report directory when no report directory is specified to "lcc run".
        """
        return create_report_dir_with_rotation(self.dir)

    def load_suites(self):
        # type: () -> Sequence[Suite]
        """
        Overridable. Load the project's suites.
        """
        return load_suites_from_directory(osp.join(self.dir, "suites"))

    def load_fixtures(self):
        # type: () -> Sequence[Fixture]
        """
        Overridable. Load the project's fixtures.
        """
        fixtures_dir = osp.join(self.dir, "fixtures")
        if osp.exists(fixtures_dir):
            return load_fixtures_from_directory(fixtures_dir)
        else:
            return []

    def pre_run(self, cli_args, report_dir):
        # type: (Any, str) -> None
        """
        Overridable. This hook is called before running the tests.
        """
        pass

    def post_run(self, cli_args, report_dir):
        # type: (Any, str) -> None
        """
        Overridable. This hook is called after running the tests.
        """
        pass

    def build_report_title(self):
        # type: () -> Optional[str]
        """
        Overridable. Build a custom report title as a string.
        """
        return None

    def build_report_info(self):
        # type: () -> List[Tuple[str, str]]
        """
        Overridable. Build a list key/value pairs (expressed as a two items tuple)
        that will be available in the report.

        Example::

            [
                ("key1", "value1"),
                ("key2", "value2")
            ]
        """
        info = []
        if self.show_command_line_in_report:
            info.append(("Command line", " ".join([os.path.basename(sys.argv[0])] + sys.argv[1:])))
        return info


def _find_file_in_parent_directories(filename, dirname):
    if osp.exists(osp.join(dirname, filename)):
        return osp.join(dirname, filename)

    parent_dirname = osp.dirname(dirname)
    if parent_dirname == dirname:
        return None  # root directory has been reached

    return _find_file_in_parent_directories(filename, parent_dirname)


def find_project_file():
    # type: () -> Optional[str]
    filename = os.environ.get("LCC_PROJECT_FILE")
    if filename is not None:
        return filename if osp.exists(filename) else None

    return _find_file_in_parent_directories(PROJECT_CONFIG_FILE, os.getcwd())


def find_project_dir():
    # type: () -> Optional[str]
    project_filename = find_project_file()
    if project_filename is None:
        return None
    return osp.dirname(project_filename)


def create_project(project_dir):
    # type: (str) -> None
    shutil.copyfile(get_resource_path(("project", "template.py")), osp.join(project_dir, PROJECT_CONFIG_FILE))
    os.mkdir(osp.join(project_dir, "suites"))
    os.mkdir(osp.join(project_dir, "fixtures"))


def load_project_from_file(project_filename):
    # type: (str) -> Project

    # load project module
    try:
        project_module = import_module(project_filename)
    except UserError as e:
        raise e  # propagate UserError
    except Exception:
        raise ProjectError("Got an unexpected error while loading project:%s" % (
            serialize_current_exception()
        ))
    
    # get project config instance
    try:
        project = project_module.project
    except AttributeError:
        raise ProjectError("Cannot find symbol 'project' in module '%s'" % project_filename)
    if not isinstance(project, Project):
        raise ProjectError("Symbol 'project' in module '%s' does not inherit lemoncheesecake.project.Project" % project_filename)
    
    return project


def load_project_from_dir(project_dir):
    # type: (str) -> Project
    return load_project_from_file(osp.join(project_dir, PROJECT_CONFIG_FILE))


def load_project():
    # type: () -> Project
    project_filename = find_project_file()
    if project_filename is None:
        raise ProjectError("Cannot find project file")

    return load_project_from_file(project_filename)



from .matchers import *
from .operations import check_that, require_that, assert_that, check_that_in, require_that_in, assert_that_in

DISPLAY_DETAILS_WHEN_EQUAL = True

# for pydoc & sphinx
__all__ = [sym_name for sym_name in dir() if not sym_name.startswith("_")]
